#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2023/12/11 16:40:25
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''


from abc import abstractmethod
from typing import List
from xagents.config import *
from xagents.kb.common import Chunk



class AbstractLoader:
    @abstractmethod
    def load(self, file_path: str) -> List[Chunk]:
        raise NotImplementedError



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2023/12/08 16:57:23
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''

import copy
from typing import Type
from langchain.vectorstores.faiss import FAISS
import faiss
from langchain.vectorstores.utils import DistanceStrategy
from langchain.vectorstores import VectorStore
from langchain.docstore.in_memory import InMemoryDocstore
from langchain.vectorstores.elasticsearch import ElasticsearchStore
from langchain.vectorstores import VectorStore as VectorStore

from xagents.model.core import EMBD

from loguru import logger

class XVecStore(VectorStore):
    @classmethod
    def is_local(cls):
        raise NotImplementedError()

    @classmethod
    def need_embd(cls):
        raise NotImplementedError()

    @classmethod
    def from_config(cls, config: dict):
        raise NotImplementedError()


class XFAISS(XVecStore, FAISS):
    @classmethod
    def is_local(cls):
        return True

    @classmethod
    def need_embd(cls):
        return False

    @classmethod
    def from_config(cls, config: dict):

        distance_strategy = config["distance_strategy"]
        embedding: EMBD = config["embedding"]
        dim_len = embedding.get_dim()

        # faiss = dependable_faiss_import()
        if distance_strategy == DistanceStrategy.MAX_INNER_PRODUCT:
            index = faiss.IndexFlatIP(dim_len)
        else:
            # Default to L2, currently other metric types not initialized.
            index = faiss.IndexFlatL2(dim_len)
        config.update(embedding_function=config.pop("embedding"))
        config.update(index=index, docstore=InMemoryDocstore(), index_to_docstore_id={})
        # logger.debug(f"{config=}")

        vecstore = cls(
            **config
        )
        return vecstore


class XES(XVecStore, ElasticsearchStore):
    @classmethod
    def is_local(cls):
        return False

    @classmethod
    def need_embd(cls):
        return True

    @classmethod
    def from_config(cls, config: dict):
        vecstore = cls(
            **config
        )
        return vecstore


_vecstores = [XFAISS, XES]
_name2vecstores = {e.__name__:e for e in _vecstores}


def list_vecstores():
    return [e.__name__ for e in _vecstores]


def get_vecstore_cls(name: str) -> Type[XVecStore]:
    return _name2vecstores[name]


def get_vector_store(config: dict, embd_model: EMBD = None) -> XVecStore:
    tmp_config = copy.copy(config)
    vs_cls = tmp_config.pop("vs_cls")
    vs_cls = get_vecstore_cls(vs_cls)
    logger.debug(f"getting vecstore with config:{config}")
    # if vs_cls.need_embd():
    #     if embd_model is None:
    #         raise ValueError("Need embd model to create vector store")
    tmp_config.update(embedding=embd_model)
    return vs_cls.from_documents([], **tmp_config)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2023/12/12 11:34:29
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''

from typing import List
from loguru import logger
from xagents.config import *
from xagents.kb.core import KnowledgeBase, KnowledgeBaseFile, get_config_path
from xagents.kb.loader import _EXT2LOADER
from xagents.kb.vector_store import list_vecstores as lv
from xagents.kb.splitter import _SPLITTERS
from langchain.vectorstores.utils import DistanceStrategy



def list_vecstores():
    return lv()


def list_distance_strategy():
    return [e.name for e in DistanceStrategy]


def list_knowledge_base_names()->List[str]:
    """列出所有知识库名称

    Returns:
        str: 知识库名称列表
    """
    kb_names = os.listdir(KNOWLEDGE_BASE_DIR)
    return kb_names

def list_knowledge_bases()->List[KnowledgeBase]:
    """列出所有知识库
    Returns:
        List[KnowledgeBase]: 知识库列表
    """
    
    kb_names = os.listdir(KNOWLEDGE_BASE_DIR)
    kbs = []
    for name in kb_names:
        config_path = get_config_path(name)
        kb = KnowledgeBase.from_config(config=config_path)
        kbs.append(kb)
    return kbs    


def get_knowledge_base(name: str) -> KnowledgeBase:
    """
    根据名称获取知识库实例
    """
    logger.debug(f"get knowledge base with name:{name}")
    kb_names = list_knowledge_base_names()
    if name not in kb_names:
        msg =  f"kb_name {name} not in {kb_names}"
        logger.warning(msg)
        raise ValueError(msg)
    
    config_path = get_config_path(name)
    kb = KnowledgeBase.from_config(config=config_path)
    return kb




def list_kb_files(kb_name: str) -> List[KnowledgeBaseFile]:
    kb = get_knowledge_base(kb_name)
    kb_files = kb.list_files()
    return kb_files

def get_knowledge_base_file(kb_name: str, file_name: str)->KnowledgeBaseFile:
    kb_files = list_kb_files(kb_name)
    kb_file_dict = {f.file_name: f for f in kb_files}
    assert file_name in kb_file_dict, f"{file_name} not in {kb_files}"
    return kb_file_dict[file_name]


def list_valid_exts():
    return list(_EXT2LOADER.keys())


def list_splitters() -> List[str]:
    return list(e.__name__ for e in _SPLITTERS)


def create_knowledge_base(name: str, desc: str,
                          embedding_config: dict, vecstore_config: dict,
                          distance_strategy: DistanceStrategy = DistanceStrategy.MAX_INNER_PRODUCT
                          ):
    """
    创建知识库
    """
    logger.info(f"Creating knowledge base {name}...")
    kb_names = list_knowledge_base_names()
    if name in kb_names:
        msg = f"Knowledge base {name} already exists, can not create!"
        logger.warning(msg)
        raise ValueError(msg)
    kb = KnowledgeBase(name=name, embedding_config=embedding_config,
                       description=desc, vecstore_config=vecstore_config, distance_strategy=distance_strategy)
    kb.save()
    return kb


def update_knowledge_base(name: str, config:dict):
    """
    创建知识库
    """
    kb = get_knowledge_base(name)
    kb.update_info(config)
    return kb





if __name__ == "__main__":
    print(list_knowledge_base_names())
    print(list_vecstores())
    print(list_distance_strategy())



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2023/12/11 14:28:24
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''

from multiprocessing import Process

def format_prompt(template: str, **kwargs):
    for k, v in kwargs.items():
        template = template.replace("{k}", str(v))
    return template




#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2024/03/11 16:05:05
@Author  :   ChenHao
@Description  :   离线任务class
@Contact :   jerrychen1990@gmail.com
'''

from random import random
import sqlite3
from enum import Enum
from time import sleep
from typing import Callable
import uuid
from loguru import logger
from pydantic import BaseModel, Field
from snippets import jdumps,jloads

from xagents.config import DB_DIR

db_file = f"{DB_DIR}/job.db"


# logger.add(sys.stdout, level="DEBUG")
class JobStatus(str, Enum):
    PENDING="PENDING"
    RUNNING="RUNNING"
    SUCCESS="SUCCESS"
    FAILED="FAILED"
    

class JobInfo(BaseModel):
    job_id:str=Field(default=str(uuid.uuid4()), description="任务ID")
    status:JobStatus=Field(default=JobStatus.PENDING, description="任务状态")
    resp:dict=Field(default={}, description="任务结果")
    func_name:str=Field(description="函数名称")
    

class Job:
    def __init__(self, func:Callable, job_id:str, status:JobStatus,resp:dict):
        self.job_id:str = job_id
        self.status:JobStatus = status
        self.func=func
        self.resp=resp
        self.save()
    
    @classmethod
    def create_job(cls, func:Callable):
        job_info = JobInfo(func_name=func.__name__)
        job = Job(func=func, **job_info.model_dump(exclude={"func_name"}))
        return job
        
    def run(self, *args, **kwargs):
        logger.info(f"{self.job_id} start running")
        self.status = JobStatus.RUNNING
        self.save()
        
        try:
            self.resp = self.func(*args, **kwargs)
            self.status=JobStatus.SUCCESS
        except Exception as e:
            logger.error(f"Job {self.job_id} failed with error {e}")
            self.resp = e
            self.status=JobStatus.FAILED
        self.save()
        logger.info(f"{self.job_id} finished")
        
    def save(self):
        sql = f"""
        insert into JOB(ID, STATUS, RESP, FUNC_NAME)
        values('{self.job_id}', '{self.status}', '{jdumps(self.resp)}', '{self.func.__name__}')
        on CONFLICT(ID) do UPDATE SET
        ID='{self.job_id}', STATUS='{self.status}', RESP='{jdumps(self.resp)}'
        """
        # logger.info(f"{sql=}")
        logger.debug(f"updating job with sql:{sql}")
        con = sqlite3.connect(db_file)
        cur = con.cursor()
        cur.execute(sql)
        con.commit()

    
        
def get_job_info(job_id:str)->JobInfo:
    sql = f"select ID,STATUS, RESP, FUNC_NAME from JOB where ID='{job_id}'"
    logger.debug(f"executing sql:{sql}")
    con = sqlite3.connect(db_file)
    cur = con.cursor()
    jobs = cur.execute(sql).fetchall()
    if not jobs:
        logger.warning(f"no job found with id {job_id}")
    else:
        item = jobs[0]
        # item[-2]=jloads(item[-2])
        # logger.info(item)
        kwargs = dict(zip(["job_id", "status", "resp", "func_name"], item))
        kwargs["resp"]=jloads(kwargs["resp"])
        job_info=JobInfo(**kwargs)

        return job_info
    
        
if __name__ == "__main__":
    con = sqlite3.connect(db_file)
    cur = con.cursor()
    # cur.execute("drop table JOB")
    
    
#     cur.execute("""
# CREATE TABLE JOB (
#     ID VARCHAR(255) PRIMARY KEY,
#     STATUS VARCHAR(255) NOT NULL,
#     RESP VARCHAR(1024),
#     FUNC_NAME VARCHAR(1024)
# )""")
                
    
    import time
    from concurrent.futures import ThreadPoolExecutor,wait,ALL_COMPLETED,FIRST_COMPLETED, as_completed
    
    
    
    
    def func(a):
        wt = random.randrange(1, 5)
        # print(wt)
        sleep(wt)
        return dict(rs=a**2)
    
    
    
    
    # lists=[4,5,2,3]
    lists = [1]
    jobs = [Job(func) for i in lists]

    # 创建一个最大容纳数量为2的线程池
    pool= ThreadPoolExecutor(max_workers=1)

    # 通过submit提交执行的函数到线程池中
    all_task=[pool.submit(j.run, i) for j,i in zip(jobs,lists)]


    # # # 通过result来获取返回值
    result=[i.result() for i in all_task]
    print(f"result:{result}")


    print("----complete-----")
    # # 线程池关闭
    pool.shutdown()
        
    
    
    
    
    



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2023/12/11 16:29:39
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''

import os
from typing import List, Type
from xagents.kb.common import Chunk
from xagents.kb.loader.common import AbstractLoader
from xagents.kb.loader.markdown import MarkDownLoader
from xagents.kb.loader.pdf import PDFLoader
from xagents.kb.loader.structed import StructedLoader

_EXT2LOADER = {
    "pdf": PDFLoader,
    "markdown": MarkDownLoader,
    "md": MarkDownLoader,
    "json": StructedLoader,
    "jsonl": StructedLoader,
    "csv": StructedLoader,
    "txt": MarkDownLoader
}


def get_loader_cls(file_path: str) -> Type[AbstractLoader]:
    return get_loader(file_path=file_path)

def get_loader(file_path:str)->AbstractLoader:
    ext = os.path.splitext(file_path)[-1].lower().replace(".", "")
    loader = _EXT2LOADER[ext]
    return loader




def load_file(file_path: str, **kwargs) -> List[Chunk]:
    loader_cls = get_loader_cls(file_path)
    loader: AbstractLoader = loader_cls(**kwargs)
    contents = loader.load(file_path)
    return contents



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2024/03/07 16:17:27
@Author  :   ChenHao
@Description  :   
@Contact :   jerrychen1990@gmail.com
'''
import os
from typing import List
import requests
from requests.auth import HTTPBasicAuth
from loguru import logger
from xagents.kb.common import DistanceStrategy


class XClient:
    def __init__(self, host: str, username: str, password: str, port=8001):
        self.host = host
        self.port = port
        self.auth = HTTPBasicAuth(username, password)

    def call_service(self, path: str, method='POST', **kwargs) -> dict:
        try:
            url = f" http://{self.host}:{self.port}/{path}"
            if method == "POST":
                resp = requests.post(url=url,  auth=self.auth, **kwargs)
            else:
                resp = requests.get(url=url, auth=self.auth, **kwargs, )
            logger.debug(f"request to {url} with {method=}, {kwargs}")
            resp.raise_for_status()
            resp = resp.json()
            code = resp["code"]
            msg = resp["msg"]
            resp_data = resp["data"]
            logger.debug(f"get resp:{code=}, {msg=}, {resp_data=}")
            return resp_data

        except Exception as e:
            logger.error(f"call service {url}failed\n error:{e}\nresp:{resp.text}")
            raise e

    def health(self):
        return self.call_service(path="health", method="GET")

    def get_job(self, job_id: str):
        return self.call_service(path=f"job", params=dict(job_id=job_id), method="GET")

    def list_kb(self):
        return self.call_service(path="kb/list", method="GET")

    def get_kb(self, kb_name: str):
        return self.call_service(path=f"kb", params=dict(kb_name=kb_name), method="GET")

    def create_kb(self, kb_name: str, description: str = None, vecstore_config: dict = dict(vs_cls='XFAISS'),
                  embedding_config: dict = dict(model_cls="ZhipuEmbedding", batch_size=16, norm=True),
                  distance_strategy=DistanceStrategy.MAX_INNER_PRODUCT):
        json_data = dict(kb_name=kb_name, description=description, distance_strategy=distance_strategy,
                         vecstore_config=vecstore_config, embedding_config=embedding_config)
        return self.call_service(path="kb/create", json=json_data)

    def update_kb(self, kb_name: str, kb_config: dict):
        json_data = dict(kb_name=kb_name, kb_config=kb_config)
        return self.call_service(path="kb/update", json=json_data)

    def delete_kb(self, kb_name: str):
        data = dict(kb_name=kb_name)
        return self.call_service(path="kb/delete", data=data)

    def build_index(self, kb_name: str):
        data = dict(kb_name=kb_name)
        return self.call_service(path="kb/build_index", data=data)

    def list_kb_files(self, kb_name: str):
        data = dict(kb_name=kb_name)
        return self.call_service(path="kb_file/list", data=data, method="GET")

    def create_kb_file(self, kb_name: str, file_path: str):
        file_name = os.path.basename(file_path)
        with open(file_path, "r") as f:
            files = {'file': (file_path, f)}
            data = dict(kb_name=kb_name, file_name=file_name)
            return self.call_service(path="kb_file/create", data=data, files=files)

    def delete_kb_file(self, kb_name: str, file_name: str):
        data = dict(kb_name=kb_name, file_name=file_name)
        return self.call_service(path="kb_file/delete", data=data)

    def get_kb_file_chunks(self, kb_name: str, file_name: str):
        data = dict(kb_name=kb_name, file_name=file_name)
        return self.call_service(path="kb_file/chunks", data=data, method="GET")

    def cut_kb_file(self,
                    kb_name: str,
                    file_name: str,
                    splitter_cls: str = "BaseSplitter",
                    splitter: str = "\n",
                    max_len: int = 200,
                    min_len: int = 10,
                    parse_table: bool = False):
        data = dict(kb_name=kb_name, file_name=file_name, splitter_cls=splitter_cls,
                    splitter=splitter, max_len=max_len, min_len=min_len, parse_table=parse_table)
        return self.call_service(path="kb_file/cut", data=data)

    def search_kb(self,
                  kb_name: str,
                  query: str,
                  file_names: List[str] = None,
                  top_k: int = 3,
                  score_threshold: float = 0.0,
                  pre_content_len: int = 0,
                  next_content_len: int = 0
                  ):
        data = dict(kb_name=kb_name, query=query, file_names=file_names, top_k=top_k, score_threshold=score_threshold,
                    pre_content_len=pre_content_len, next_content_len=next_content_len)
        return self.call_service(path="kb/search", data=data)
